''' Tests for ProprietaryPeriodSetter '''

import logging
import random
import subprocess

import pytest
from astropy.time import Time
from proprietary_setter.project_finders import \
    LegacyProjectFinder, NewArchiveProjectFinder
from proprietary_setter.prop_setter import \
    MAX_DURATION, make_parser, ProprietaryPeriodSetter, ProjectLock
from schema import Project, create_session
from schema.legacy_model import LegacyProject
from support.logging import get_console_logger

# pylint: disable=logging-format-interpolation

_PROFILE = 'local'

_LOG = get_console_logger('proprietary_setter_tests')

canonical_values = {'13B-014': {'duration': 365,
                                'lock': ProjectLock.UTIME.value,
                                'expire': 365},
                    'BG266': {'duration': 365,
                              'lock': ProjectLock.LTIME.value,
                              'expire': 0},
                    'BZ076': {'duration': 548,
                              'lock': ProjectLock.LTIME.value,
                              'expire': 0},

                    }

def test_setter_updates_vla_project():
    project_code = '13B-014'
    legacy_before = LegacyProjectFinder(_PROFILE).find_projects(project_code)[0]
    new_duration = random.randint(0, MAX_DURATION)

    expected_proprietary = legacy_before.stoptime + new_duration
    expected_lock = compute_lock(expected_proprietary)

    args = ['-P', _PROFILE, project_code, str(new_duration)]
    parser = make_parser()
    namespace = parser.parse_args(args)
    usage = parser.format_usage()

    # call the setter
    setter = ProprietaryPeriodSetter(usage, namespace)
    projects = setter.set_project_proprietary_state()
    assert len(projects) == 2
    assert 'legacy' in projects.keys()

    # check that we got what we expected from the legacy DB
    legacy_project = projects['legacy'][0]
    assert legacy_project is not None
    assert isinstance(legacy_project, LegacyProject)
    assert legacy_project.project_lock == expected_lock
    assert legacy_project.proprietary == expected_proprietary
    assert new_duration == legacy_project.proprietary_duration

    # check that we got what we expected from the new-archive DB
    assert 'new' in projects.keys()
    project_after = projects['new']
    assert project_after is not None
    assert isinstance(project_after, Project)

    assert project_after.proprietary_duration == new_duration
    _restore_original_values(project_code)

def test_invalid_duration_is_rejected():
    project_code = '13B-014'
    new_duration = MAX_DURATION + 1
    args = ['-P', _PROFILE, project_code, str(new_duration)]
    parser = make_parser()

    namespace = parser.parse_args(args)
    usage = parser.format_usage()

    with pytest.raises(SystemExit) as exc:
        ProprietaryPeriodSetter(usage, namespace)
        if not exc.value.code:
            _restore_original_values(project_code)
        assert exc.value.code == 2

    new_duration = -1
    args = ['-P', _PROFILE, project_code, str(new_duration)]
    parser = make_parser()
    namespace = parser.parse_args(args)
    with pytest.raises(SystemExit) as exc:
        ProprietaryPeriodSetter(usage, namespace) \
            .set_project_proprietary_state()
        if not exc.value.code:
            _restore_original_values(project_code)
        assert exc.value.code == 2

def test_invalid_project_type_is_rejected():
    ''' script should complain and exit if project is not VLA/EVLA'''
    project_codes =['BZ076', 'AGBT18A_314', '2019.1.01832.S']
    new_duration = MAX_DURATION - 10

    for project_code in project_codes:
        args = ['-P', _PROFILE, project_code, str(new_duration)]
        parser = make_parser()
        namespace = parser.parse_args(args)
        with pytest.raises(SystemExit) as exc:
            ProprietaryPeriodSetter(parser.format_usage(), namespace) \
                .set_project_proprietary_state()
            if not exc.value.code:
                _restore_original_values(project_code)
        assert exc.value.code == 3

def test_bad_project_code_causes_update_failure():
    project_code = '15B-99C'
    new_duration = MAX_DURATION - 10
    args = ['-P', _PROFILE, project_code, str(new_duration)]
    parser = make_parser()
    namespace = parser.parse_args(args)
    with pytest.raises(SystemExit) as exc:
        ProprietaryPeriodSetter(parser.format_usage(), namespace) \
            .set_project_proprietary_state()
        if not exc.value.code:
            _restore_original_values(project_code)
        assert exc.value.code == 3

def test_cli_returns_expected_codes():
    project_code = '13B-014'
    new_duration = str(random.randint(0, MAX_DURATION))

    # the "control": good arguments
    args = ['-P', _PROFILE, project_code, new_duration]
    return_code = CommandLineLauncher().run(args)
    if not return_code:
        _restore_original_values(project_code)
    assert return_code == 0

    # profile value omitted; parser will think project_code is profile
    args = ['-P', project_code, new_duration]
    _LOG.info('>>> profile value omitted')
    return_code = CommandLineLauncher().run(args)
    if not return_code:
        _restore_original_values(project_code)
    assert return_code == 2

    # profile argument omitted entirely
    _LOG.info('>>> profile argument omitted')
    args = [project_code, new_duration]
    return_code = CommandLineLauncher().run(args)
    if not return_code:
        _restore_original_values(project_code)
    assert return_code == 1

    # invalid profile: prop_setter will be unable to connect to new archive DB
    _LOG.info('>>> invalid profile')
    args = ['-P', 'foo', project_code, new_duration]
    return_code = CommandLineLauncher().run(args)
    if not return_code:
        _restore_original_values(project_code)
    assert return_code == 3

    # project code omitted
    _LOG.info('>>> project code omitted')
    args = ['-P', _PROFILE, new_duration]
    return_code = CommandLineLauncher().run(args)
    if not return_code:
        _restore_original_values(project_code)
    assert return_code == 2

    # invalid proprietary duration
    _LOG.info('>>> invalid duration')
    args = ['-P', _PROFILE, project_code, '999']
    return_code = CommandLineLauncher().run(args)
    if not return_code:
        _restore_original_values(project_code)
    assert return_code == 2

    # proprietary duration omitted
    _LOG.info('>>> duration omitted')
    args = ['-P', _PROFILE, project_code]
    return_code = CommandLineLauncher().run(args)
    if not return_code:
        _restore_original_values(project_code)
    assert return_code == 2

    # nonexistent project code
    _LOG.info('>>> bad project code')
    args = ['-P', _PROFILE, 'foo', new_duration]
    return_code = CommandLineLauncher().run(args)
    if not return_code:
        _restore_original_values(project_code)
    assert return_code == 3

    # VLBA project
    _LOG.info('>>> VLBA project (not permitted)')
    args = ['-P', _PROFILE, 'BG266', new_duration]
    return_code = CommandLineLauncher().run(args)
    if not return_code:
        _restore_original_values(project_code)
    assert return_code == 3


### UTILITIES ###

def compute_lock(proprietary):
    ''' For the legacy archive only:
        decide the value of project.project_lock
        based on value of project.proprietary
        and the current datetime.
    '''
    return ProjectLock.LTIME.value if proprietary > Time.now().mjd \
        else ProjectLock.UTIME.value

def _restore_original_values(project_code: str) -> None:
    values_dict = canonical_values[project_code]

    # set canonical values for this project....

    # ...in the legacy archive
    legacy_projs = LegacyProjectFinder(_PROFILE) \
        .find_projects(project_code)

    legacy_session = create_session('LEGACY', profile=_PROFILE)
    try:
        for legacy_proj in legacy_projs:
            legacy_proj = legacy_session.query(LegacyProject) \
                .filter(LegacyProject.project_code
                        .in_((project_code, legacy_proj.project_code))) \
                .first()
            assert legacy_proj is not None
            legacy_proj.proprietary_duration = values_dict['duration']
            legacy_proj.unlock_expire = values_dict['expire']
            legacy_proj.project_lock = values_dict['lock']
        legacy_session.commit()

    except Exception as exc:
        pytest.fail(f'problem updating {project_code} in legacy archive:'
                    f' {exc}')

    finally:
        legacy_session.close()

    # ...in the new archive
    session = create_session('SDM', profile=_PROFILE)
    try:
        new_proj = session.query(Project) \
            .filter(Project.project_code == project_code) \
            .first()
        new_proj.proprietary_duration = values_dict['duration']
        session.commit()
    except Exception as exc:
        pytest.fail(f'problem updating {project_code} in new archive:'
                    f' {exc}')
    finally:
        session.close()

    # confirm values have been reset in the legacy archive....
    legacies = LegacyProjectFinder(_PROFILE).find_projects(project_code)
    for legacy_proj in legacies:
        assert values_dict['duration'] == legacy_proj.proprietary_duration
        assert values_dict['lock'] == legacy_proj.project_lock
        assert values_dict['expire'] == legacy_proj.unlock_expire

    # ... and in the "new" archive
    newbie = NewArchiveProjectFinder(_PROFILE).find_project(project_code)
    assert values_dict['duration'] == newbie.proprietary_duration


class CommandLineLauncher:
    ''' Simulates execution of the prop setter CLI'''

    def run(self, args_in):
        ''' Launch prop setter with args_in '''

        args = ['proj_prop_period']
        for arg in args_in:
            args.append(arg)

        try:
            # N.B. deliberately -NOT- using a context manager ("with") here
            # to ensure we always get a return code
            proc = subprocess.run(args,
                                  stdout=subprocess.PIPE,
                                  stderr=subprocess.STDOUT,
                                  check=False,
                                  bufsize=1,
                                  universal_newlines=True)

            if not proc.returncode:
                if proc.returncode == 0:
                    return 0

                # (should never happen)
                _LOG.error(f'No returncode for {args}')
                return -1
            else:
                logging.error(proc.stdout)
                return proc.returncode

        except subprocess.CalledProcessError as exc:
            _LOG.error(f'{exc}')
            return exc.returncode
