"""
Model classes used by the Workspaces system.
"""

import enum
import json
from pathlib import Path
from typing import Dict, List

import sqlalchemy as sa
from pycapo import CapoConfig
from sqlalchemy import create_engine
from sqlalchemy.orm import relationship, sessionmaker
from sqlalchemy.ext.declarative import declarative_base


class AbstractFile:
    """
    Abstract file is exactly that, an abstract concept of what a file is, to be
    returned from various non-filesystem places.
    """
    filename: str
    content: bytes

    def __init__(self, filename: str, content: bytes):
        self.filename, self.content = filename, content

    def write_to(self, directory: Path):
        (directory / self.filename).write_bytes(self.content)

    def __json__(self, request):
        return {'filename': self.filename, 'content': self.content}

    @classmethod
    def from_path(cls, path: Path) -> 'AbstractFile':
        return cls(path.name, path.read_bytes())


class WorkflowEventType(enum.Enum):
    """
    Kinds of events that a workflow can send. These four are derived from
    the many kinds that HTCondor can send.
    """
    SUBMITTED = 0
    EXECUTING = 1
    TERMINATED = 5
    OTHER = -1


class WorkflowEvent:
    """
    An event from the underlying workflow execution system, which is likely
    to be HTCondor.
    """
    def __init__(
            self,
            job_name: str,
            job_id: str,
            event_type: WorkflowEventType,
            timestamp: str,
            log: str,
            retval: int = None):
        # NOTE: when you modify the properties here, also update the schema
        # in the .json package.
        self.job_name = job_name
        self.job_id = str(job_id)
        self.type = event_type
        self.timestamp = timestamp
        self.log = log
        self.retval = retval

    def json(self) -> str:
        d = {
            'job_name': self.job_name,
            'type': str(self.type),
            'timestamp': self.timestamp,
            'return_value': self.retval,
            'log': self.log
        }
        return json.dumps(d)

    def __str__(self):
        return f'WorkflowEvent with data {self.json()}'

    def __repr__(self):
        return f'<WorkflowEvent {self.__dict__}>'

    def __eq__(self, other):
        return self.job_name == other.job_name and \
               self.job_id == other.job_id and \
               self.type == other.type and \
               self.timestamp == other.timestamp and \
               self.log == other.log and \
               self.retval == other.retval



Base = declarative_base()


class Workflow(Base):
    """
    A Workflow is a suite of tasks defined by templates that must be
    executed together with some user-submitted files.
    """
    __tablename__ = 'workflows'
    workflow_name = sa.Column('workflow_name', sa.String, primary_key=True)
    templates = relationship('WorkflowTemplate', backref='workflow')
    requests = relationship('WorkflowRequest', backref='workflow')

    def __json__(self, request):
        return {'workflow_name': self.workflow_name, 'templates': self.templates}

    def __repr__(self):
        return f'<Workflow workflow_name={self.workflow_name}>'

    def render_templates(self, argument: Dict, files: List[Path]) -> List[AbstractFile]:
        """
        Render the templates associated with this workflow
        :param argument: the workflow argument JSON
        :param files:    the files to be processed
        :return:         a list of rendered templates
        """
        raise NotImplementedError


class WorkflowTemplate(Base):
    """
    Workflow Templates are Mustache-formatted files associated with a given Workflow
    """
    __tablename__ = 'workflow_templates'
    filename = sa.Column('filename', sa.String, primary_key=True)
    content = sa.Column('content', sa.LargeBinary, nullable=False)
    workflow_name = sa.Column('workflow_name', sa.String, sa.ForeignKey('workflows.workflow_name'), primary_key=True)

    def render(self, argument: Dict) -> AbstractFile:
        raise NotImplementedError('Add mustache to render this template')

    @property
    def template_file(self) -> AbstractFile:
        return AbstractFile(self.filename, self.content)

    @template_file.setter
    def template_file(self, file: AbstractFile):
        self.filename, self.content = file.filename, file.content

    def __json__(self, request):
        # Defer to the actual template file's contents for JSON conversion
        return self.template_file.__json__(request)

    def __repr__(self):
        return f'<WorkflowTemplate filename={self.filename} for workflow={self.workflow_name}>'


class WorkflowRequest(Base):
    """
    Workflow Requests invoke a particular workflow with a particular argument and some user-supplied files.
    """
    __tablename__ = 'workflow_requests'
    workflow_request_id = sa.Column('workflow_request_id', sa.Integer, primary_key=True, autoincrement=True)
    workflow_name = sa.Column('workflow_name', sa.String, sa.ForeignKey('workflows.workflow_name'))
    argument = sa.Column('argument', sa.JSON)
    files = relationship('WorkflowRequestFile', backref='request')


class WorkflowRequestFile(Base):
    """
    A Workflow Request File is a file supplied by the user and attached to the request they have submitted.
    """
    __tablename__ = 'workflow_request_files'
    workflow_request_id = sa.Column('workflow_request_id', sa.Integer,
                                    sa.ForeignKey('workflow_requests.workflow_request_id'),
                                    primary_key=True)
    filename = sa.Column('filename', sa.String, primary_key=True)
    content = sa.Column('content', sa.LargeBinary, nullable=False)

    @property
    def file(self) -> AbstractFile:
        return AbstractFile(self.filename, self.content)

    @file.setter
    def set_file(self, file: AbstractFile):
        self.filename, self.content = file.filename, file.content

    def __repr__(self):
        return f'<WorkflowRequestFile filename={self.filename}>'


def get_engine():
    """
    Generate the SQL Alchemy engine for us, using Capo.
    :return:
    """
    capo = CapoConfig().settings('metadataDatabase')
    url = capo.jdbcUrl.replace('jdbc:', '').replace('://', f'://{capo.jdbcUsername}:{capo.jdbcPassword}@')
    return create_engine(url)


def get_session_factory(engine):
    """
    Generate the SQL Alchemy session factory for us, using the supplied engine
    :param engine:
    :return:
    """
    factory = sessionmaker()
    factory.configure(bind=engine)
    return factory
